import pytest

import sheraf.databases
from .utils import (
    close_database,
    create_blob_directory,
    create_database,
    create_temp_directory,
    delete_temp_directory,
    start_postgresql_server,
    start_zeo_server,
    stop_postgresql_server,
    stop_zeo_server,
)


def load_database_kwargs(request, default=None):
    if not request.cls:
        return default or {}

    db_kwargs = default or {}

    if hasattr(request.cls, "sheraf_database_kwargs"):
        db_kwargs = request.cls.sheraf_database_kwargs

    if hasattr(request.cls, "sheraf_database_extra_kwargs"):
        db_kwargs.update(request.cls.sheraf_database_extra_kwargs)

    return db_kwargs


@pytest.fixture
def sheraf_temp_dir():
    """Creates a temporary directory to store sheraf data.

    :return: The directory path
    """
    temp_directory, old_files_root_dir = create_temp_directory()
    try:
        yield temp_directory
    finally:
        delete_temp_directory(temp_directory, old_files_root_dir)


@pytest.fixture
def sheraf_database(request):
    """This fixture creates a database.

    If a test needs to pass arguments to the database creation, it must be a
    method of a class defining ``sheraf_database_kwargs`` or/and
    ``sheraf_database_extra_kwargs``.
    If ``sheraf_database_kwargs`` is defined, it will overwrite the default
    (empty) arguments dictionnary. If ``sheraf_database_extra_kwargs`` defined,
    the arguments dictionnary will be updates by its values.

    By default, a in-memory database is created with
    :class:`~ZODB.DemoStorage.DemoStorage`.

    >>> def test_something(sheraf_database):
    ...     with sheraf.connection(commit=True):
    ...         MyModel.create()
    ...
    >>> class TestSomething:
    ...     sheraf_database_extra_kwargs = {"database_name": "foobar_db"}
    ...
    ...     def test_something(self, sheraf_database):
    ...         assert "foobar_db" == sheraf_database.name
    """

    db_kwargs = load_database_kwargs(request)
    temp_directory, old_files_root_dir = create_temp_directory()
    database = None
    try:
        database = create_database(db_kwargs)
        yield database

    finally:
        close_database(database)
        delete_temp_directory(temp_directory, old_files_root_dir)


@pytest.fixture
def sheraf_connection(sheraf_database):
    """This fixture opens a connection on a database.

    >>> def test_something(sheraf_connection):
    ...     MyModel.create()
    """
    connection = sheraf_database.connection_open()
    try:
        yield connection
    finally:
        connection.transaction_manager.abort()
        sheraf_database.connection_close(connection)


@pytest.fixture
def sheraf_filestorage_database(request):
    """This fixture creates a database using a
    :class:`~ZODB.FileStorage.FileStorage`."""

    temp_directory, old_files_root_dir = create_temp_directory()
    db_kwargs = load_database_kwargs(
        request, {"uri": "file:///{}/Data.fs".format(temp_directory)}
    )

    database = None
    try:
        database = create_database(db_kwargs)
        yield database

    finally:
        close_database(database)
        delete_temp_directory(temp_directory, old_files_root_dir)


@pytest.fixture(scope="session")
def sheraf_zeo_server():
    """This fixture launches a ZEO server."""
    temp_directory, old_files_root_dir = create_temp_directory()
    blob_dir = create_blob_directory(temp_directory)
    zeo_process, zeo_port = start_zeo_server(temp_directory, blob_dir)
    yield zeo_port, blob_dir
    stop_zeo_server(zeo_process)
    delete_temp_directory(temp_directory, old_files_root_dir)


@pytest.fixture
def sheraf_zeo_database(request, sheraf_zeo_server):
    """This fixture launches a ZEO server, and then creates a database using a
    :class:`~ZEO.ClientStorage`.

    .. note :: Only one ZEO server is launched by test session.
    """

    db_kwargs = load_database_kwargs(
        request,
        {
            "uri": "zeo://localhost:{}?blob_dir={}&shared_blob_dir=true".format(
                *sheraf_zeo_server
            )
        },
    )

    database = None
    try:
        database = create_database(db_kwargs)
        yield database

    finally:
        if database:
            with sheraf.databases.connection(commit=True) as conn:
                for key in list(conn.root().keys()):
                    del conn.root()[key]
        close_database(database)


@pytest.fixture(scope="session")
def sheraf_pgsql_server():
    """This fixture launches a PostgreSQL server, if available."""

    temp_directory, old_files_root_dir = create_temp_directory()
    pg_process, pg_user, pg_port, pg_error = start_postgresql_server(temp_directory)
    if pg_error:
        pytest.skip(pg_error)
    yield pg_user, pg_port
    stop_postgresql_server(pg_process)
    delete_temp_directory(temp_directory, old_files_root_dir)


@pytest.fixture
def sheraf_pgsql_relstorage_database(request, sheraf_pgsql_server):
    """This fixture launches a PostgreSQL server, and then creates a database
    using a :class:`~relstorage.storage.RelStorage`.

    .. note :: Only one PostgreSQL server is launched by test session.
    """

    db_kwargs = load_database_kwargs(
        request, {"uri": "postgres://{}@localhost:{}/zodb".format(*sheraf_pgsql_server)}
    )

    database = None
    try:
        database = create_database(db_kwargs)
        yield database

    finally:
        if database:
            with sheraf.databases.connection(commit=True) as conn:
                for key in list(conn.root().keys()):
                    del conn.root()[key]
        close_database(database)



import io
import os
import shutil
import stat
import subprocess
import sys
import tempfile
import time
import warnings

import portpicker
import ZODB.POSException

import sheraf
from sheraf.databases import Database, LocalData


class Conflict(object):
    def __call__(self, times=1):
        if not hasattr(self, "times"):
            self.times = 0
            raise ZODB.POSException.ConflictError()
        else:
            self.times = self.times + 1
            if self.times < times:
                raise ZODB.POSException.ConflictError()

    def reset(self):
        try:
            delattr(self, "times")
        except AttributeError:
            pass


conflict = Conflict()


def create_temp_directory():
    persistent_dir_path = tempfile.mkdtemp()
    _old_files_root_dir = sheraf.attributes.files.FILES_ROOT_DIR
    sheraf.attributes.files.FILES_ROOT_DIR = os.path.join(persistent_dir_path, "files/")
    os.makedirs(sheraf.attributes.files.FILES_ROOT_DIR)
    return persistent_dir_path, _old_files_root_dir


def delete_temp_directory(persistent_dir_path, _old_files_root_dir):
    shutil.rmtree(persistent_dir_path)
    sheraf.attributes.files.FILES_ROOT_DIR = _old_files_root_dir


def create_blob_directory(persistent_dir):
    blob_dir = os.path.join(persistent_dir, "blobs/")
    os.makedirs(blob_dir)
    os.chmod(blob_dir, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
    return blob_dir


def create_database(db_args):
    try:
        return Database(**db_args)
    except KeyError:  # pragma: no cover
        db_name = db_args.get("database_name") or Database.DEFAULT_DATABASE_NAME
        last_context = LocalData.get().last_database_context.get(db_name)
        warnings.warn(
            "Database '{}' was already open on {} at line {}".format(
                db_name, last_context[0], last_context[1]
            )
            if last_context
            else "Database '{}' was already open".format(db_name)
        )
        Database.get(db_name).close()
        return Database(**db_args)


def close_database(database=None):
    if database:
        database.close()

    for db_name, db in list(Database.all()):
        last_context = LocalData.get().last_database_context.get(db.name)
        warnings.warn(
            "Database '{}' was not closed. It should be closed inside the test. It was opened on {} at {}".format(
                db_name, last_context[0], last_context[1]
            )
            if last_context
            else "Database '{}' was not closed. It should be closed inside the test.".format(
                db_name
            )
        )
        db.close()


ZEO_CONFIGURATION_FILE_CONTENT = """
<zeo>
    address localhost:{port}
</zeo>
<filestorage>
    path {path}
    {blob_dir}
</filestorage>
<eventlog>
    level DEBUG
    <logfile>
        path STDOUT
        format %(asctime)s %(levelname)s %(message)s
    </logfile>
</eventlog>
"""


def create_zeo_configuration_file(
    zeo_port, zeo_conf_path, file_storage_path, blob_dir=None
):
    conf_content = ZEO_CONFIGURATION_FILE_CONTENT.format(
        port=zeo_port,
        path=file_storage_path,
        blob_dir="blob-dir {}".format(blob_dir) if blob_dir else "",
    )
    with io.open(zeo_conf_path, "w") as f:
        f.write(conf_content)


def start_zeo_server(persistent_dir, blob_dir=None):
    zeo_port = portpicker.pick_unused_port()
    file_storage_path = os.path.join(persistent_dir, "Data.fs")
    zeo_conf_path = os.path.join(persistent_dir, "zeo.conf")

    create_zeo_configuration_file(zeo_port, zeo_conf_path, file_storage_path, blob_dir)
    zeo_process = subprocess.Popen(
        ("env", "PYTHONPATH=.", "runzeo", "-C", zeo_conf_path),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    return zeo_process, zeo_port


def stop_zeo_server(zeo_process, silent=False):
    zeo_process.kill()
    stdout, stderr = zeo_process.communicate()

    if stdout and not silent:
        print("[runzeo STDOUT]")
        sys.stdout.write(stdout.decode())

    if stderr and not silent:  # pragma: no cover
        print("[runzeo STDERR]")
        sys.stdout.write(stderr.decode())


def start_postgresql_server(persistent_dir):
    AFTER_PSQL_SERVER_LAUNCH_SLEEP_TIME = 1
    pg_user = os.getlogin()
    pg_port = portpicker.pick_unused_port()
    pg_dir = os.path.join(persistent_dir, "pgdata")
    try:
        exit_code = subprocess.call(
            ("initdb", "--auth", "trust", "--pgdata", pg_dir, "--username", pg_user),
            stdout=open("/dev/null"),
        )
    except IOError:  # pragma: no cover
        warnings.warn(
            "Postgresql seems to not be installed on this system. Be sure 'initdb' is in your $PATH"
        )
        return None, None, None, "Postgresql seems to not be installed on this system."

    if exit_code == 1:  # pragma: no cover
        return None, None, None, "Postgresql server initialization has failed."

    try:
        pg_process = subprocess.Popen(
            (
                "postgres",
                "-D",
                pg_dir,
                "-p",
                "{}".format(pg_port),
                "-k",
                persistent_dir,
            ),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except IOError:  # pragma: no cover
        warnings.warn(
            "Postgresql seems to not be installed on this system. Be sure 'postgres' is in your $PATH"
        )
        return None, None, None, "Postgresql seems to not be installed on this system."

    time.sleep(AFTER_PSQL_SERVER_LAUNCH_SLEEP_TIME)

    try:
        exit_code = subprocess.call(
            ("createdb", "zodb", "--port", "{}".format(pg_port), "--host", "localhost"),
            stdout=open("/dev/null"),
        )
    except IOError:  # pragma: no cover
        warnings.warn(
            "Postgresql seems to not be installed on this system. Be sure 'createdb' is in your $PATH"
        )
        return None, None, None, "Postgresql seems to not be installed on this system."

    if exit_code == 1:  # pragma: no cover
        return None, None, None, "Postgresql database creation has failed."

    return pg_process, pg_user, pg_port, None


def stop_postgresql_server(pg_process, silent=False):
    pg_process.kill()
    stdout, stderr = pg_process.communicate()

    if stdout and not silent:  # pragma: no cover
        print("[postgresql STDOUT]")
        sys.stdout.write(stdout.decode(errors="ignore"))

    if stderr and not silent:  # pragma: no cover
        print("[postgresql STDERR]")
        sys.stdout.write(stderr.decode(errors="ignore"))
